"""
Base Correction class
"""

from abc import ABC, abstractmethod
import os
from astropy.io import fits
from pykoala import VerboseMixin, vprint

class CorrectionBase(ABC, VerboseMixin):
    """
    Abstract base class for implementing astronomical corrections on a 
    DataContainer.

    This class defines the structure and basic functionality required for 
    applying a correction to an astronomical data container. It includes
    logging capabilities to track the application of corrections.

    Attributes
    ----------
    verbose : bool
        Convenient attribute for controlling the logging output. If True, logs are set to 
        INFO level; if False, logs are set to ERROR level.
    logger : logging.Logger
        Logger instance used to log messages. If not provided, a pykoala logger child
        (see `pykoala.ancillary.pykoala_logger`) is created with a name based on the
        correction's name.
    log_filename : str, optional
        Name of the file to log messages. If provided, logs will also be saved 
        to this file.
    log_level : int, optional
        Level of logging, which overrides the `verbose` attribute if provided
        (for more details see `logging levels <https://docs.python.org/3/library/logging.html#logging-levels>`_.

    Methods
    -------
    name :
        Abstract property that needs to be defined in a subclass, representing 
        the name of the correction.
    verbose :
        Abstract property that needs to be defined in a subclass, representing 
        the verbosity level of the correction.
    apply :
        Abstract method that must be implemented in a subclass to apply the 
        correction.
    vprint(msg, level='info') :
        Logs a message at the specified level.
    record_correction(datacontainer, status='applied', **extra_comments) :
        Logs the status of the correction in the DataContainer, with additional
        information if provided.

    Examples
    --------
    To create a new correction, subclass CorrectionBase and implement the 
    ``name``, ``verbose``, and ``apply`` methods.

    Notes
    -----
    When implementing a new correction, ensure that the ``apply`` method correctly
    applies the intended transformation to the DataContainer and that the 
    operation is logged via the ``record_correction`` method.

    """

    def __init__(self, verbose=True, logger=None, log_filename=None, log_level=None) -> None:

        if logger is None:
            self.logger = f"correction.{self.name}"
        else:
            self.logger = logger

        self.verbose = verbose
        # This variable supersedes `verbose`
        if log_level is not None:
            self.logger.setLevel(log_level)

        if log_filename is not None:
            self.log_into_file(log_filename, level=log_level)

    @property
    @abstractmethod
    def name(self):
        """Name of the correction."""
        return None

    @abstractmethod
    def apply(self):
        raise NotImplementedError("Each class needs to implement the `apply` method")

    def record_correction(self, datacontainer, status='applied', **extra_comments):
        """
        Logs the status of the correction in the DataContainer.

        Whenever a correction is applied, this method logs the status of the 
        correction (e.g., 'applied' or 'failed') and any additional comments in 
        the DataContainer's log.

        Parameters
        ----------
        datacontainer : :class:`pykoala.data_container.DataContainer`
            The data container to log the correction.
        status : str, optional
            Indicates the success of the correction. Should be either 'applied' 
            or 'failed' (default is 'applied').
        **extra_comments : dict
            Additional information to log alongside the correction status.

        Raises
        ------
        KeyError
            If `status` is not 'applied' or 'failed'.
        """
        if status not in ['applied', 'failed']:
            raise KeyError("Correction log status can only be 'applied' or 'failed'")
        
        datacontainer.history(self.name, status, tag='correction')
        for k, v in extra_comments.items():
            datacontainer.history(self.name, f"{k} {v}", tag='correction')


class CorrectionOffset(object):
    """Relative Offset Data.

    This class stores a 2D relative offset.

    Attributes
    ----------
    offset_data : np.ndarray
        offset correction data.
    offset_error : np.ndarray
        Standard deviation of ``offset_data``.
    path: str
        Filename path.

    """
    offset_data = None
    offset_error = None

    def __init__(self, path=None, offset_data=None, offset_error=None):
        self.path = path
        self.offset_data = offset_data
        self.offset_error = offset_error

    def tofits(self, output_path=None):
        """Save the offset in a FITS file.
        
        Parameters
        ----------
        output_path: str, optional, default=None
            FITS file name path. If None, and ``self.path`` exists,
            the original file is overwritten.

        Notes
        -----
        The output fits file contains an empty PrimaryHDU, and two ImageHDU
        ("OFFSET", "OFFSET_ERR") containing the offset data and associated error.
        """
        if output_path is None:
            if self.path is None:
                raise NameError("Provide output path")
            else:
                output_path = self.path
        primary = fits.PrimaryHDU()
        data = fits.ImageHDU(data=self.offset_data, name='OFFSET')
        error = fits.ImageHDU(data=self.offset_error, name='OFFSET_ERR')
        hdul = fits.HDUList([primary, data, error])
        hdul.writeto(output_path, overwrite=True)
        hdul.close(verbose=True)
        vprint(f"{self.__class__.__name__} data saved at {output_path}")

    @classmethod
    def from_fits(cls, path):
        """Load the offset data from a fits file.

        Loads offset values (extension 1) and
        associated errors (extension 2) from a fits file.

        Parameters
        ----------
        path : str
            Path to the FITS file containing the offset data.

        Returns
        -------
        wavelength_offset : :class:`WavelengthOffset`
            A :class:`WavelengthOffset` initialised with the input data.
        """
        if not os.path.isfile(path):
            raise NameError(f"offset file {path} does not exist.")
        vprint(f"Loading wavelength offset from {path}")
        with fits.open(path) as hdul:
            offset_data = hdul[1].data
            offset_error = hdul[2].data
        return cls(offset_data=offset_data, offset_error=offset_error,
                   path=path)



from matplotlib import pyplot as plt
from matplotlib import colors
from matplotlib.ticker import AutoMinorLocator
from matplotlib.collections import PatchCollection
from matplotlib.cm import ScalarMappable

import numpy as np

from astropy import units as u
from astropy import visualization
from astropy.visualization import quantity_support

from pykoala import vprint
from pykoala import ancillary

quantity_support()
# plt.style.use('dark_background')
SYMMETRIC_CMAP = plt.get_cmap('seismic').copy()
SYMMETRIC_CMAP.set_extremes(bad='gray', under='cyan', over='fuchsia')

DEFAULT_CMAP = plt.get_cmap("gist_earth").copy()
DEFAULT_CMAP.set_bad('gray')

def local_quantity_support(func):
    """Allow astropy Quantities support locally."""
    def wrapper(*args, **kwargs):
        with quantity_support():
            return func(*args, **kwargs)
    return wrapper
        

def default_ax_setting(ax):
    ax.xaxis.set_minor_locator(AutoMinorLocator())
    ax.yaxis.set_minor_locator(AutoMinorLocator())
    ax.tick_params(which='both', bottom=True,
                   top=True, left=True, right=True)
    ax.tick_params(which='major', direction='inout',
                   length=8, grid_alpha=.3)
    ax.tick_params(which='minor', direction='in', length=2, grid_alpha=.1)
    ax.grid(True, which='both')

def new_figure(fig_name,
               tweak_axes=True,
               figsize=None,
               **kwargs):
    """
    Close old version of the figure and create new one.
    

    Parameters
    ----------
    figsize : tuple
        Figure size.
    nrows : int
        Number of rows.
    ncols : int
        Number of columns.
    sharex : str or bool, optional
        Whether panels share the x axis. True, False, or 'col'
        (default; x-axis shared accross columns)
    sharey : str or bool, optional
        Whether panels share the y axis. True, False, or 'row'
        (default; y-axis shared accross rows)
    **gridspec_kw : dict, optional
        Default sets height and width space to `{'hspace': 0, 'wspace': 0})`

    Returns
    -------
    fig : plt.Figure
    axes : ndarray(mpl.Axes)
    """

    plt.close(fig_name)

    if "sharex" not in kwargs:
        kwargs["sharex"] = "col"
    if "sharex" not in kwargs:
        kwargs["sharex"] = "row"
    if "gridspec_kw" not in kwargs:
        kwargs["gridspec_kw"] = {'hspace': 0, 'wspace': 0}
    if "squeeze" not in kwargs:
        kwargs["squeeze"] = False

    if figsize is None:
        figsize = (9 + kwargs.get("ncols", 1),
                   4 + kwargs.get("nrows", 1))

    fig, axes = plt.subplots(num=fig_name, figsize=figsize,
                             layout="constrained", **kwargs)
    if tweak_axes:
        for ax in axes.flat:
            default_ax_setting(ax)
    fig.suptitle(fig_name)

    return fig, axes

def plot_image(fig, ax, data,
               cmap=DEFAULT_CMAP,
               xlabel=None, x=None,
               ylabel=None, y=None,
               cbax=None, norm=None,
               cblabel=None,
               norm_interval=visualization.AsymmetricPercentileInterval,
               interval_args={"lower_percentile": 1.0,
                              "upper_percentile": 99.0},
               stretch=visualization.PowerStretch, stretch_args={"a": 0.7}):
    """
    Plot a colour map (imshow) with axes and colour scales.

    Parameters
    ----------
    fig : plt.Figure
        Figure where the colour map will be drawn.
    ax : mpl.Axes
        Axes where the colour map will be drawn.
    data : ndarray
        2D array to be represented.
    cmap : str or mpl.colors.Norm
    xlabel: str, optional
        Label of x axis.
    x : ndarray, optional
        Values along x axes (defaults to pixel number)
    ylabel : str, optional
        Label of y axis.
    y : ndarray, optional
        Values along y axes (defaults to pixel number)
    cbax: mpl.Axes
        Axes where the colour bar will be drawn.
    norm : mpl.colors.Norm
    cblabel : str
        Label of the colorbar
    Returns
    -------
    im : mpl.AxesImage
    cb : mpl.Colorbar
    """

    if y is None:
        y = np.arange(data.shape[0])
    if x is None:
        x = np.arange(data.shape[1])

    if isinstance(data, u.Quantity):
        unit = data.unit
        value = data.value
    else:
        value = data
        unit = None

    if norm is None:
        interval = norm_interval(**interval_args)
        norm = visualization.ImageNormalize(value, interval=interval,
                              stretch=stretch(**stretch_args),
                              clip=False)
    elif isinstance(norm, str):
        norm = getattr(colors, norm)()

    im = ax.imshow(value,
                   extent=(x[0]-(x[1]-x[0])/2, x[-1]+(x[-1]-x[-2])/2,
                           y[0]-(y[1]-y[0])/2, y[-1]+(y[-1]-y[-2])/2),
                   interpolation='none', origin='lower',
                   cmap=cmap,
                   norm=norm,
                   )
    ax.set_aspect('auto')
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if cbax is None:
        cb = fig.colorbar(im, ax=ax, orientation='vertical', shrink=.9,
                          extend='both')
        cbax = cb.ax
    elif cbax:
        cb = fig.colorbar(im, cax=cbax, orientation='vertical', extend='both')
    else:
        cb = None
    if cbax:
        cb.ax.yaxis.set_label_position("left")
        if unit is not None:
            if cblabel is None:
                cblabel = f"{unit}"
        cb.set_label(cblabel)
        cb.ax.tick_params(labelsize='small')
    return im, cb

def plot_fibres(fig, ax, rss=None, x=None, y=None,
                fibre_diam=None, data=None, 
                patch_args={}, use_wcs=False, fix_limits=True,
                cmap=DEFAULT_CMAP, norm=None, cbax=None, cblabel=None, 
                norm_interval=visualization.MinMaxInterval, interval_args={},
                stretch=visualization.LinearStretch, stretch_args={}):
    """
    Plot a colour map of a physical magnitude defined on each fibre.

    Parameters
    ----------
    fig : plt.Figure
        Figure where the colour map will be drawn.
    ax : mpl.Axes
        Axes where the colour map will be drawn.
    rss : RSS, optional, default=None
        Row-Stacked Spectra containing the fibre positions.
    x : np.ndarray or astropy.units.Quantity, optional, default=None
        Fibre position values along the x axis.
    y : np.ndarray or astropy.units.Quantity, optional, default=None
        Fibre position values along the y axis.
    fibre_diam : float or astropy.units.Quantity, optional, default=1.25 arcsec
        Fibre diameter.
    data : ndarray, optional, default=None
        1D array to be represented. If None, fibres will appear as empty circles.
    patch_args : dict, optional
        Additional arguments passed to each fibre :class:`plt.Circle` patch.
    use_wcs : bool, optional, default=False
        If True, use axes WCS world transformation for plotting the patches.
    fix_limits : bool, optional, default=True
        If True, set the limits of the axes using the edge values of ``x`` and ``y``.
    cmap : str or mpl.colors.Norm, optional
        Colormap used to plot the values of ``data``.
    norm : str or mpl.colors.Norm, optional
        Normalization map for plotting ``data``.
    cbax: mpl.Axes
        Axes where the plt.Colorbar will be drawn.
    clabel: str
        Colorbar label.
    norm_interval : astropy.visualization.BaseInterval, optional, default=MinMaxInterval
        Interval to create a normalization map.
    interval_args : dict
        Additional arguments to be passed to the interval.
    stretch : astropy.visualization.BaseStretch, optional, default=LinearStretch
        Stretching used on the normalization map.
    stretch_args : dict, optional
        Additional arguments to be passed to stretch.
    Returns
    -------
    ax : mpl.Axes
    patch_collection : mpl.PatchCollection
    cb : mpl.Colorbar
    """
    if rss is not None:
        x = rss.info['fib_ra']
        y = rss.info['fib_dec']
        fibre_diam = rss.fibre_diameter
    else:
        if fibre_diam is None:
            raise ValueError("Must provide a fibre diameter value")

    if not isinstance(x, u.Quantity):
        # Assume that the values of x and y are sky positions
        x = x << u.degree
        y = y << u.degree

    if fix_limits:
        ax.set_xlim((x.min() - fibre_diam).value,
                    (x.max() + fibre_diam).value)
        ax.set_ylim((y.min() - fibre_diam).value,
                    (y.max() + fibre_diam).value)

    if norm is None:
        interval = norm_interval(**interval_args)
        norm = visualization.ImageNormalize(data, interval=interval,
                              stretch=stretch(**stretch_args),
                              clip=False)
    elif isinstance(norm, str):
        norm = getattr(colors, norm)()

    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap).copy()

    if data is not None:
        fib_colors = cmap(norm(data))
        mappable = ScalarMappable(norm=norm, cmap=cmap)
        if cbax is None:
            cb = fig.colorbar(mappable, ax=ax, orientation='vertical', shrink=.9)
            cbax = cb.ax
        elif cbax is False:
            cb = None
        else:
            cb = fig.colorbar(mappable, cax=cbax, orientation='vertical')
        if cbax:
            cb.ax.yaxis.set_label_position("left")
            cb.set_label(cblabel)
            cb.ax.tick_params(labelsize='small')
    else:
        fib_colors = ["none"] * len(x)
        cb = None

    if "edgecolor" not in patch_args:
        patch_args["edgecolor"] = "k"

    if use_wcs:
        patch_args["transform"] = ax.get_transform('world')

    patches = [plt.Circle(
        xy=(x_c, y_c), radius=fibre_diam.to_value(x.unit) / 2, facecolor=color,
        **patch_args
        ) for x_c, y_c, color in zip(x.value, y.value, fib_colors)]
    patch_collection = PatchCollection(patches, match_original=True,
                                       label='Fibre')
    ax.add_collection(patch_collection)
    return ax, patch_collection, cb

def qc_cube(cube, spax_pct=[75, 90, 99]):
    """Create a quality control (QC) plot for a Cube.

    Parameters
    ----------
    cube: pykoala.data_container.Cube
        The cube containing the data to plot.

    spax_pct: array_like (default: [75, 90, 99])
        Spaxel ranks. Number of spanxels per dimension. 

    Returns
    -------
    figure: matplotlib.Figure
    """

    fig = plt.figure(figsize=(12, 12))
    vprint(f"[QCPLOT] Cube QC plot for: {cube.info['name']}")
    plt.suptitle(cube.info['name'])
    gs = fig.add_gridspec(5, 4, wspace=0.35, hspace=0.25)

    # Maps -----
    wl_spaxel_idx = np.sort(np.random.randint(
        low=0, high=cube.intensity.shape[0], size=3))
    wl_col = ['b', 'g', 'r']
    for wl_idx, i in zip(wl_spaxel_idx, range(3)):
        ax = fig.add_subplot(gs[0, i:i+1])
        default_ax_setting(ax)
        ax.set_title(r"$\lambda@${:.1f}".format(cube.wavelength[wl_idx]),
                     fontdict=dict(color=wl_col[i]))
        ax, cb = plot_image(fig, ax, data=cube.intensity[wl_idx],
                            cblabel="Intensity", cmap="cividis")
        cb.ax.yaxis.set_label_position("right")
        ax = fig.add_subplot(gs[1, i:i+1])
        default_ax_setting(ax)
        ax, cb = plot_image(fig, ax, 
                   data=cube.intensity[wl_idx] / cube.variance[wl_idx]**0.5,
                   cblabel="SNR",
                   cmap="jet")
        cb.ax.yaxis.set_label_position("right")
    
    # Plot the mean intensity
    mean_intensity = np.nanmedian(cube.intensity, axis=0)
    mean_variance = np.nanmedian(cube.variance, axis=0)
    mean_intensity[~np.isfinite(mean_intensity)] = 0
    mean_instensity_pos = np.argsort(mean_intensity.flatten())
    mapax = fig.add_subplot(gs[1, -1])
    default_ax_setting(mapax)
    _, cb = plot_image(fig, mapax, 
                   data=mean_intensity / mean_variance**0.5,
                   cblabel="SNR",
                   cmap="jet")
    cb.ax.yaxis.set_label_position("right")

    mapax = fig.add_subplot(gs[0, -1])
    mapax.set_title("Median")
    _, cb = plot_image(fig, mapax,  data=mean_intensity, cblabel="Intensity",
                       cmap="cividis")
    cb.ax.yaxis.set_label_position("right")
    # ------ Spectra -------
    pos_col = ['purple', 'orange', 'cyan']
    x_spaxel_idx = np.random.randint(low=0,
                                     high=cube.intensity.shape[1],
                                     size=3)
    y_spaxel_idx = np.random.randint(low=0,
                                     high=cube.intensity.shape[2],
                                     size=3)
    spaxel_entries = mean_instensity_pos[
        np.array(mean_instensity_pos.size / 100 * np.array(spax_pct),
                 dtype=int)]
    x_spaxel_idx, y_spaxel_idx = np.unravel_index(spaxel_entries,
                                                  shape=mean_intensity.shape)
    ax = fig.add_subplot(gs[2:3, :])
    default_ax_setting(ax)
    for x_idx, y_idx, i in zip(x_spaxel_idx, y_spaxel_idx, range(3)):
        ax.plot(cube.wavelength, cube.intensity[:, x_idx, y_idx], lw=0.8,
                color=pos_col[i])
        mapax.plot(y_idx, x_idx, marker='+', ms=8, mew=2, lw=2, color=pos_col[i])
    for i, wl in enumerate(cube.wavelength[wl_spaxel_idx]):
        ax.axvline(wl, color=wl_col[i], zorder=-1, alpha=0.8)
    ax.axhline(0, alpha=0.2, color='r')
    ylim = np.nanpercentile(
        cube.intensity[np.isfinite(cube.intensity)].value, [40, 95])
    ylim[1] *= 20
    ylim[0] *= 0.1
    np.clip(ylim, a_min=0, a_max=None, out=ylim)
    ax.set_ylim(ylim)
    ax.set_yscale('symlog', linthresh=0.1)
    ax.set_ylabel(f"Flux ({cube.intensity.unit})")

    # SNR ------------
    ax = fig.add_subplot(gs[3:5, :], sharex=ax)
    default_ax_setting(ax)
    for x_idx, y_idx, i in zip(x_spaxel_idx, y_spaxel_idx, range(3)):
        ax.plot(cube.wavelength,
                cube.intensity[:, x_idx, y_idx] / cube.variance[:, x_idx, y_idx]**0.5, lw=0.8,
                color=pos_col[i],
                label=f"Spaxel rank={spax_pct[i]}")
    ax.legend()
    ax.set_ylabel("SNR/pix")
    ax.set_xlabel(f"Wavelength ({cube.wavelength.unit})")
    plt.close(fig)
    return fig

@ancillary.remove_units_dec
def qc_cubing(rss_weight_maps, exposure_times):
    """..."""
    if exposure_times.ndim == 1:
        exposure_times = (rss_weight_maps
        * exposure_times[:, np.newaxis, np.newaxis, np.newaxis])
    n_rss = rss_weight_maps.shape[0]

    p5, p95 = np.nanpercentile(rss_weight_maps, [.1 , 99.9])
    w_im_args = dict(vmin=p5 * 0.9, vmax=p95 * 1.1, cmap='nipy_spectral',
                     interpolation='none', origin='lower', aspect="auto")

    tp5, tp95 = np.nanpercentile(exposure_times, [5 , 95])
    print("Mean exposure time (s): ", np.nanmean(exposure_times))
    t_im_args = dict(vmin=tp5 * 1.1, vmax=tp95 * 1.1, cmap='gnuplot',
                     interpolation='none', origin='lower', aspect="auto")
    fig, axs = plt.subplots(ncols=n_rss + 1, nrows=2, sharex=True,
     sharey=True, constrained_layout=True, gridspec_kw={"hspace": 0.05})
    for i in range(n_rss):
        ax = axs[0, i]
        ax.set_title(f"RSS - {i + 1}")
        ax.imshow(np.nanmedian(rss_weight_maps[i], axis=0), **w_im_args)
        ax = axs[1, i]
        ax.imshow(np.nanmedian(exposure_times[i], axis=0), **t_im_args)

    axs[0, -1].set_title("Mean")
    mappable = axs[0, -1].imshow(np.nanmedian(np.nanmean(rss_weight_maps, axis=0), axis=0),
                      **w_im_args)
    plt.colorbar(mappable, ax=axs[0, -1], label='Median weight')
    mappable = axs[1, -1].imshow(np.nanmedian(np.nanmean(exposure_times, axis=0), axis=0),
                      **t_im_args)
    plt.colorbar(mappable, ax=axs[1, -1], label='Median exp. time (s) / pixel')
    plt.close()
    return fig

@ancillary.remove_units_dec
def qc_fibres_on_fov(fov_size, pixel_colum_pos, pixel_row_pos,
                     fibre_diam=1.25):
    fig = plt.figure(constrained_layout=True)
    ax = fig.add_subplot(111)
    ax.set_xlim(np.min([-10, pixel_colum_pos.min() - 5]),
                np.max([fov_size[1] + 10, pixel_colum_pos.max() + 5]))
    ax.set_ylim(np.min([-10, pixel_row_pos.min() - 5]),
                np.max([fov_size[0] + 10, pixel_row_pos.max() + 5]))
    ax.set_xlabel("Column pixel (RA)")
    ax.set_ylabel("Row pixel (DEC)")
    ax.grid(visible=True, alpha=0.3)
    
    patches = [plt.Circle(xy=(c, r), radius=fibre_diam/2) for c, r in zip(
        pixel_colum_pos, pixel_row_pos)]
    p = PatchCollection(patches, facecolors='tomato', edgecolors='k', label='Fibre')
    ax.add_collection(p)
    patch = plt.Rectangle(xy=(-.5, -.5),
                          width=fov_size[1] + 0.5,
                          height=fov_size[0] + 0.5,
                          fc='none', ec='k', lw=2, label='Cube FoV')
    ax.add_patch(patch)
    ax.legend(handles=[patch])
    return fig
# =============================================================================
# Star profile
# =============================================================================


def qc_moffat(intensity, ra_offset, dec_offset, fit_model):
    """#TODO"""

    r = np.sqrt(
        (ra_offset - fit_model.x_0.value)**2 + (dec_offset - fit_model.y_0.value)**2)
    r = r.flatten()
    I = intensity.flatten()
    I_hat = fit_model(ra_offset, dec_offset).flatten()

    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111)
    ax.set_title("2D Moffat fit")

    inax = ax.inset_axes((0, 0.65, 1, 0.35))
    inax.plot(r, np.abs(I - I_hat) / I, 'k+')
    inax.grid(visible=True)
    inax.set_ylabel(r'$\frac{I-\hat{I}}{I}$', fontsize=17)
    inax.set_xlabel(r'$|r-\hat{r}_0|$ (arcsec)', fontsize=15)
    inax.set_ylim(-0.099, 2)
    # Input data
    inax = ax.inset_axes((0, 0.0, 0.5, 0.5))
    c = inax.pcolormesh(ra_offset, dec_offset,
                        np.log10(intensity), cmap='nipy_spectral')
    inax.plot(fit_model.x_0.value, fit_model.y_0.value, 'k+', ms=14, mew=2,
              label=f'centre ({fit_model.x_0.value:.1f}, {fit_model.y_0.value:.1f}) arcsec')
    plt.colorbar(mappable=c, ax=inax, orientation='horizontal', anchor=(0, -1),
                 label=r"$\log_{10}(I)$")
    inax.legend()
    inax.set_ylabel("DEC offset (arcsec)")
    inax.set_xlabel("RA offset (arcsec)")
    
    inax = ax.inset_axes((0.55, 0.0, 0.5, 0.5))
    c = inax.pcolormesh(ra_offset, dec_offset,
                        np.log10(intensity/I_hat.reshape(intensity.shape)),
                    vmin=-.5, vmax=.5, cmap='seismic')
    inax.plot(fit_model.x_0.value, fit_model.y_0.value, 'k+', ms=14, mew=2)
    plt.colorbar(mappable=c, ax=inax, orientation='horizontal', anchor=(0, -1),
                 label=r"$\log_{10}(I/\hat{I})$")
    inax.set_ylabel("DEC offset (arcsec)")
    inax.set_xlabel("RA offset (arcsec)")

    ax.axis("off")

    return fig


# =============================================================================
# Registration
# =============================================================================

def qc_registration(rss_list, **kwargs):
    n_rss = len(rss_list)
    fig, axs = plt.subplots(nrows=1, ncols=n_rss+1,
                            figsize=(4*(n_rss + 1), 4),
                            gridspec_kw=dict(hspace=0.50))
    axs[0].set_title("Input RSS overlap")
    axs[0].set_xlabel("RA (deg)")
    axs[0].set_ylabel("DEC (deg)")

    cmap = plt.get_cmap('jet', n_rss)
    for i, rss in enumerate(rss_list):
        axs[0].scatter(rss.info['ori_fib_ra'] / 3600 + rss.info['ori_cen_ra'],
                       rss.info['ori_fib_dec'] / 3600 + rss.info['ori_cen_dec'],
                       marker='o', ec=cmap(i / (n_rss - 1)), c='none',
                   #alpha=1/n_rss
                   )
        axs[i+1].set_title(f"Re-centered RSS-{i+1}" + "\n"
                           f"(ra, dec) shift: {rss.info['cen_ra'] * 3600:.2f}, {rss.info['cen_dec'] * 3600:.2f} ('')")
        axs[i+1].scatter(rss.info['fib_ra'],
                   rss.info['fib_dec'],
                   c=np.nansum(rss.intensity, axis=1),
                   norm=colors.LogNorm(),
                   marker='o', cmap='Greys_r',
                   )
        axs[i+1].axvline(0, c=cmap(i / (n_rss - 1)), lw=1.5)
        axs[i + 1].axhline(0, c=cmap(i / (n_rss - 1)), lw=1.5)
        axs[i + 1].set_xlabel("RA Offset (arcsec)")
        axs[i + 1].set_ylabel("DEC Offset (arcsec)")

    return fig

def qc_registration_crosscorr(images_list, cross_corr_results):
    """TODO..."""
    # Set reference points to illustrate the shifts

    vmin, vmax = np.nanpercentile(images_list, [5, 95])
    imargs = dict(vmin=vmin, vmax=vmax, cmap='viridis', interpolation='none')

    fig, axs = plt.subplots(nrows=1, ncols=len(images_list),
                          figsize=(4 * len(images_list), 4))
    plt.suptitle("QC cross-correlation-based image registration")
    
    axs[0].set_title("Reference image")
    mappable = axs[0].imshow(images_list[0], **imargs)

    # Reference image centre will be used as visual reference
    centre = (images_list[0].shape[1] / 2, images_list[0].shape[0] / 2)
    axs[0].plot(*centre, 'r+')
    # Plot the rest of images
    for i, im in enumerate(images_list[1:]):
        shift = cross_corr_results[i][0]
        ax = axs[i + 1]
        mappable = ax.imshow(images_list[i + 1], **imargs)
        ax.arrow(*centre, - shift[1], - shift[0], color='r', width=1)
        ax.plot(centre[0] - shift[1], centre[1] - shift[0])
        ax.annotate(f"Shif (row, col): {shift[0]:.2f}, {shift[1]:.2f}", xy=(0.05, 0.95),
                    color='red',
                    xycoords='axes fraction', va='top', ha='left')
    cax = ax.inset_axes((1.05, 0, 0.05, 1))
    plt.colorbar(mappable, cax=cax)
    plt.close(fig)
    return fig

def qc_registration_centroids(images_list, wcs_list, offsets, ref_pos):
    """TODO..."""
    # Account for images with different sizes
    vmin, vmax = np.nanpercentile(
        np.hstack([im.flatten().value for im in images_list]), [5, 95])
    imargs = dict(vmin=vmin, vmax=vmax, cmap='viridis', interpolation='none')

    ncols=len(images_list)
    fig = plt.Figure(figsize=(4 * ncols, 4))
    plt.suptitle("QC centroid-based image registration")
    for i in range(ncols):
        ax = fig.add_subplot(1, ncols, i + 1 , projection=wcs_list[i])
    
        mappable = ax.imshow(images_list[i].value, **imargs)
        ax.scatter(ref_pos.ra, ref_pos.dec, marker='*',
                   ec='r', label='Reference', transform=ax.get_transform('world'))
        ax.scatter(ref_pos.ra - offsets[i][0], ref_pos.dec - offsets[i][1], marker='o',
                   ec='k', label='Centroid', transform=ax.get_transform('world'))

        ax.annotate(f"Offset (ra, dec):\n {offsets[i][0].to('arcsec').value:.2f}, {offsets[i][1].to('arcsec').value:.2f} arcsec",
                    xy=(0.05, 0.95),
                    color='red',
                    xycoords='axes fraction', va='top', ha='left')
    ax.legend(bbox_to_anchor=(0.5, 1.1), loc='lower center')
    cax = ax.inset_axes((1.05, 0, 0.05, 1))
    plt.colorbar(mappable, cax=cax)
    plt.close(fig)
    return fig

# =============================================================================
# Astrometry
# =============================================================================


def qc_external_image(ref_image, ref_wcs, external_image, external_image_wcs):
    fig = plt.figure(constrained_layout=True)
    ax = fig.add_subplot(111, projection=external_image_wcs)

    contourf_params = dict(cmap='Spectral', levels=[18, 19, 20, 21, 22, 23],
                               vmin=19, vmax=23, extend='both')
    contour_params = dict(levels=[18, 19, 20, 21, 22, 23],
                          colors='k')

    ax.coords.grid(True, color='orange', ls='solid')
    ax.coords[0].set_format_unit('deg')
    mappable = ax.contourf(external_image, **contourf_params)
    plt.colorbar(mappable, ax=ax,
                    label=r"$\rm \log_{10}(F_\nu / 3631 Jy / arcsec^2)$")
    ax.contour(ref_image,
                transform=ax.get_transform(ref_wcs), **contour_params)
    plt.close(fig)
    return fig

# =============================================================================
# Flux calibration
# =============================================================================

def qc_stellar_extraction():
    pass



"""
Module for estimating a fibre throughput correction.
"""
# =============================================================================
# Basics packages
# =============================================================================
import os
import numpy as np
import copy
from astropy.io import fits
from scipy.ndimage import median_filter, percentile_filter
# from scipy.ndimage import gaussian_filter
# =============================================================================
# Astropy and associated packages
# =============================================================================

# =============================================================================
# KOALA packages
# =============================================================================
# Modular
from pykoala import vprint
from pykoala.corrections.correction import CorrectionBase
from pykoala.data_container import RSS
from pykoala import ancillary
from pykoala.plotting import utils as plot_utils

class Throughput(object):
    """Class that represents a throughput data set.
    
    Attributes
    ----------
    throughput_data : np.ndarray
        Array containing the values of the throughput.
    throughput_error : np.ndarray
        Array containing the associated error of ``throughput_data``.
    throughput_path : str
        Path to the original file that was used to initialise the throughput.
    """
    def __init__(self, throughput_data, throughput_error,
                 throughput_file=None):
        self.throughput_data = throughput_data
        self.throughput_error = throughput_error
        self.throughput_file = throughput_file

    def to_fits(self, output_path=None):
        """Save the current throughput into a FITS file.

        Description
        -----------
        The throughput information is stored in a FITS file that consists of:

        - An empty primary HDu
        - Two ImageHDUs containing the data (extension='THROU') and associated error (extension='THROUERR') of the throughput, respectively.

        Parameters
        ----------
        output_path : str, optional, default=None
            Path to the output file where the throughput information will be stored.
            If None, and ``self.throughput_path`` is not ``None`` the original file
            will be overwritten with the new data.
        """
        if output_path is None:
            if self.throughput_file is None:
                raise NameError("Provide output file name for saving throughput")
            else:
                output_path = self.throughput_file
        primary = fits.PrimaryHDU()
        thr = fits.ImageHDU(data=self.throughput_data,
                            name='THROU')
        thr_err = fits.ImageHDU(data=self.throughput_error,
                                name='THROUERR')
        hdul = fits.HDUList([primary, thr, thr_err])
        hdul.writeto(output_path, overwrite=True)
        hdul.close(verbose=True)
        vprint(f"Throughput saved at {output_path}")

    @classmethod
    def from_fits(cls, path):
        """Create a :class:`Throughput` from an input FITS file.
        
        Throughput data and associated errors must be stored in HDUL extension 1,
        and 2, respectively.

        Parameters
        ----------
        path: str
            Path to the FITS file containing the throughput data.

        Returns
        -------
        throughput : :class:`Throughput`
            A :class:`Throughput` initialised with the input data.
        """
        if not os.path.isfile(path):
            raise NameError(f"Throughput file {path} does not exists.")
        vprint(f"Loading throughput from {path}")
        with fits.open(path) as hdul:
            throughput_data = hdul[1].data
            throughput_error = hdul[2].data
        return cls(throughput_data, throughput_error, path)

    def plot(self, pct=[1, 50, 99], random_seed=50):
        """Plot the Throughput data.
        
        Description
        -----------
        The figure is composed of four panels. The top-row panels correspond
        to a image of the Throughput values and the histogram distribution.
        The panel in the middle row displays the Throughput values along the 
        spectral axis of 3 randomly selected fibres, whereas the bottom panel
        shows the Throughput values of 3 randomly selected columns along the fibre
        direction.

        Parameters
        ----------
        pct : list, optional, default=[1, 50, 99]
            List of percentiles to compute. The first element will be used
            to symmetrize the Throughput image colour bar and histogram with
            respect to 1.
        random_seed : int, optional, default=50
            Random seed to use for generating random column/row indices.
        
        Returns
        -------
        fig : :class:`matplotlib.pyplot.Figure`
            Figure containing the plots.
        """
        fig, ax = plot_utils.new_figure(
            fig_name="throughput", figsize=(10, 8),
            ncols=1, nrows=1
            )
        ax[0, 0].axis("off")
        gs = fig.add_gridspec(3, 4, wspace=0.15, hspace=0.35)
        
        # Throughput map
        ax = fig.add_subplot(gs[0, 0:-1])
        p_values = np.nanpercentile(self.throughput_data, pct)
        im, cb = plot_utils.plot_image(fig, ax, cblabel="Throughput",
                                       xlabel="wavelength axis",
                                       ylabel="Fibre",
                                       norm=plot_utils.colors.Normalize(
                                           vmin=p_values[0],
                                           vmax=2 - p_values[0]),
                                        cmap=plot_utils.SYMMETRIC_CMAP,
                                       data=self.throughput_data)
        # Histogram
        ax = fig.add_subplot(gs[0, -1])
        ax.hist(self.throughput_data.flatten(),
                bins=self.throughput_data.size // 1000, range=[
                    p_values[0] - 0.1, 2.1 - p_values[0]],
                log=True)
        for p_name, p in zip(pct, p_values):
            ax.axvline(p, label=f"P{p_name}", ls=':')
        ax.set_ylabel("N pixels")
        ax.set_xlabel("Throughput value")
        ax.set_ylim(10, self.throughput_data.size // 100)

        ax = fig.add_subplot(gs[1, :])

        median_wavelength_throughput = np.nanmedian(self.throughput_data, axis=0)
        std_wavelength_throughput = np.nanmedian(
            np.abs(self.throughput_data - median_wavelength_throughput[np.newaxis, :]),
            axis=0) * 1.4826
        ax.fill_between(np.arange(0, self.throughput_data.shape[1]),
                        median_wavelength_throughput - std_wavelength_throughput,
                        median_wavelength_throughput + std_wavelength_throughput,
                        alpha=0.3, color='r', label='Median +/- (MAD * 1.4826)')
        ax.plot(median_wavelength_throughput, label='Median',
                lw=0.7, color='r')
        np.random.seed(random_seed)
        fibre_idx = np.random.randint(low=0, high=self.throughput_data.shape[0], size=3)
        for idx in fibre_idx:
            ax.plot(self.throughput_data[idx], label='Fibre {}'.format(idx),
                    lw=1., alpha=0.8)
        ax.set_ylim(0.75, 1.25)
        ax.set_xlabel("Spectral pixel")
        ax.legend(ncol=3)

        ax = fig.add_subplot(gs[-1, :])
        wl_idx = np.random.randint(low=0, high=self.throughput_data.shape[1],
                                   size=3)
        for idx in wl_idx:
            ax.plot(self.throughput_data[:, idx].squeeze(),
                    label='Wave column {}'.format(idx), lw=0.7,
                    alpha=1.0)
        ax.set_ylim(0.75, 1.25)
        ax.set_xlabel("Fibre number")
        ax.legend(ncol=4)
        
        return fig
        

class ThroughputCorrection(CorrectionBase):
    """
    Throughput correction class.

    This class accounts for the relative flux loss due to differences
    on the fibre efficiencies.

    Attributes
    ----------
    name : str
        Correction name, to be recorded in the log.
    throughput : :class:`Throughput`
        2D fibre throughput (n_fibres x n_wavelengths).
    verbose: bool
        False by default.
    """
    name = "ThroughputCorrection"
    throughput = None

    def __init__(self, throughput=None, throughput_path=None,
                 **correction_kwargs):
        super().__init__(**correction_kwargs)

        if throughput is None:
            self.throughput = Throughput()
        else:
            assert isinstance(throughput, Throughput)
            self.throughput = throughput

        self.throughput.path = throughput_path
        if self.throughput.throughput_data is None and self.throughput.path is not None:
            self.throughput.load_fits(self.throughput_path)

    @classmethod
    def from_file(cls, path):
        """Creates a `ThroughputCorrection` using an input FITS file.
        
        Parameters
        ----------
        path: str
            Path to the FITS file containing the throughput data.
        
        Returns
        -------
        throughput_correction : :class:`ThroughputCorrection`
            ThroughputCorrection initialised with the input Throughput data.
        """
        throughput = Throughput.from_fits(path)
        return cls(throughput, path)

    @classmethod
    def from_rss(cls, rss_set, clear_nan=True, statistic='median', medfilt=5,
                 pct_outliers=[5, 95]):
        """Compute the throughput correctoin from a set of (dome/sky)flat exposures.

        Description
        -----------
        Given a set of flat exposures, this method will estimate the average
        efficiency of each fibre.

        Parameters
        ----------
        rss_set: list
            List of RSS data.
        clean_nan: bool, optional, default=True
            If True, nan values will be replaced by a
            nearest neighbour interpolation.
        statistic: str, optional, default='median'
            Set to 'median' or 'mean' to compute the throughput function.
        medfilt: float, optional, default=None
            If provided, apply a median filter to the throughput estimate.
        pct_outliers : 2-element tupla
            Percentile limits to clip outliers.

        Returns
        -------
        throughput_correction : :class:`ThroughputCorrection`
            ThroughputCorrection initialised with the resulting throughput
            estimation.
        """
        if statistic == 'median':
            stat_func = np.nanmedian
        elif statistic == 'mean':
            stat_func = np.nanmean

        fluxes = []
        for rss in rss_set:
            f = rss.intensity / rss.info['exptime']
            f[f <= 0] = np.nan
            fluxes.append(f)
        # Combine all RSS
        combined_flux = stat_func(fluxes, axis=0)
        combined_flux_err = np.nanstd(fluxes, axis=0) / np.sqrt(len(fluxes))

        # Normalize fibres
        reference_fibre = stat_func(combined_flux, axis=0)
        throughput_data = combined_flux / reference_fibre[np.newaxis, :]

        # Estimate the error
        throughput_error = combined_flux_err / reference_fibre[np.newaxis, :]

        p5, p95 = np.nanpercentile(throughput_data, pct_outliers, axis=1)
        outliers_mask = (throughput_data - p5[:, np.newaxis] <= 0
                         ) | (throughput_data - p95[:, np.newaxis] >= 0)
        throughput_data[outliers_mask] = np.nan

        if clear_nan:
            pix = np.arange(throughput_data.shape[1])
            for ith, (f, f_err) in enumerate(zip(throughput_data, throughput_error)):
                mask = np.isfinite(f) & np.isfinite(f_err)
                throughput_data[ith] = np.interp(pix, pix[mask], f[mask])
                throughput_error[ith] = np.interp(pix, pix[mask], f_err[mask])

        if medfilt is not None:
            print(f"Applying median filter (size={medfilt} px)")
            throughput_data = median_filter(throughput_data, size=medfilt,
                                            axes=1)
            throughput_error = median_filter(
                throughput_error**2, size=medfilt)**0.5

        throughput = Throughput(throughput_data=throughput_data,
                                throughput_error=throughput_error)
        return cls(throughput)


    def apply(self, rss):
        """Apply a 2D throughput model to a RSS.

        Parameters
        ----------
        rss : :class:`pykoala.rss.RSS`
            Original Row-Stacked-Spectra to be corrected.

        Returns
        -------
        rss_corrected : :class:`pykoala.rss.RSS`
            Corrected copy of the input RSS.
        """
        assert isinstance(rss, RSS), "Throughput can only be applied to RSS data"
        rss_out = rss.copy()
        rss_out.intensity = rss_out.intensity / self.throughput.throughput_data
        rss_out.variance = rss_out.variance / self.throughput.throughput_data**2
        self.record_correction(rss_out, status='applied')
        return rss_out

# =============================================================================
# Mr Krtxo \(ﾟ▽ﾟ)/
#                                                       ... Paranoy@ Rulz! ;^D



# -*- coding: utf-8 -*-
import setuptools

import versioneer

#DESCRIPTION_FILES = ["pypi-intro.rst"]
#
#long_description = []
#import codecs
#for filename in DESCRIPTION_FILES:
#    with codecs.open(filename, 'r', 'utf-8') as f:
#        long_description.append(f.read())
#long_description = "\n".join(long_description)
setuptools.setup(
    version = versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)




"""
Atmospheric extinction and refraction effects corrections.
"""

# =============================================================================
# Basics packages
# =============================================================================
from matplotlib import pyplot as plt
import numpy as np
import os
from astropy import units as u
# =============================================================================
# KOALA packages
# =============================================================================
from pykoala import vprint
from pykoala.data_container import SpectraContainer
from pykoala.corrections.correction import CorrectionBase
from pykoala.ancillary import check_unit


class AtmosphericExtCorrection(CorrectionBase):
    r"""Atmospheric Extinction Correction.

    This module accounts for the brightness reduction caused due to the absorption of 
    photons by the atmosphere.

    For a given observed (:math:`F_{obs}`) and intrinsic flux (:math:`F_{int}`), the extinction
    :math:`E` takes the form:

    .. math::
        F_{int}(\lambda) = E(\lambda) * F_{obs}(\lambda)
        
        E(\lambda) = 10^{0.4 \cdot airmass \cdot \eta(\lambda)}

    where :math:`\eta(\lambda)` corresponds to the wavelength-dependent extinction curve.

    Attributes
    ----------
    extinction_curve: np.ndarray, optional, default=None
        Atmospheric extinction curve.
    extinction_curve_wave: np.ndarray, optional, default=None
        Atmospheric extinction curve wavelength array.
    extinction_file: str
        Path to a text file containing a wavelength, and a extinction curve.
        If None, a default extinction model will be used, corresponding to the extinction curve at Siding Spring Observatory.
    """
    name = "AtmosphericExtinction"
    verbose = True
    default_extinction = os.path.join(os.path.dirname(__file__), '..', 'input_data',
                                      'observatory_extinction', 'ssoextinct.dat')
    def __init__(self,
                 extinction_curve=None,
                 extinction_curve_wave=None,
                 extinction_curve_file='unknown',
                 **correction_args):
        super().__init__(**correction_args)
        self.vprint("Initialising correction")

        # Initialise variables
        self.extinction_curve = extinction_curve
        self.extinction_curve_wave = check_unit(
            extinction_curve_wave, u.angstrom)
        self.extinction_curve_file = extinction_curve_file

    @classmethod
    def from_text_file(cls, path=None):
        r"""Initialise the Correction from a text file.
        
        Parameters
        ----------
        path : str, optional, default=``self.default_extinction``
            Path to the file containing the extinction curve. The first and 
            second columns of the file must contain the wavelength and the value
            of :math:`\eta(\lambda)`, respectively.
        
        Returns
        -------
        correction : AtmosphericExtCorrection
            An atmospheric extinction correction.
        """
        if path is None:
            path = cls.default_extinction
        wavelength, extinct = np.loadtxt(path, unpack=True)
        return cls(extinction_curve=extinct,
                   extinction_curve_wave=wavelength << u.angstrom,
                   extinction_curve_file=path)

    def extinction(self, wavelength : u.Quantity, airmass):
        """Compute the atmospheric extinction for a given airmass and wavelength.
        
        Parameters
        ----------
        wavelength: np.ndarray
            Input array of wavelengths where to estimate the extinction.
        airmass: float
            Target airmass at which the observation is performed.
        
        Returns
        -------
        extinction: np.ndarray
            Extinction at a given wavelength and airmass.
        """
        extinction_curve = np.interp(wavelength,
                                        self.extinction_curve_wave,
                                        self.extinction_curve)
        return 10**(0.4 * airmass * extinction_curve)

    def apply(self, spectra_container, airmass=None):
        """Apply the Extinction Correction to a DataContainer.
        
        Parameters
        ----------
        airmass: float, optional
            If provided, the extinction will be computed using this value,
            otherwise the airmass stored at the `info` attribute will be used.
        
        Returns
        -------
        corrected_spectra_container: SpectraContainer
            SpectraContainer with the corrected intensity and variance.
        """
        assert isinstance(spectra_container, SpectraContainer)
        spectra_container_out = spectra_container.copy()
        if airmass is None:
            airmass = spectra_container.info['airmass']

        if self.extinction_curve is not None:
            self.vprint("Applying model-based extinction correction to"
                        f"Data Container ({airmass:.2f} airmass)")
            extinction = self.extinction(spectra_container_out.wavelength, airmass)
            comment = ("Atm. extinction file :" + os.path.basename(
                        self.extinction_curve_file)
                        + f"|airmass={airmass:.2f}")
        else:
            raise AttributeError("Extinction correction not provided")

        # Apply the correction
        extinction = np.expand_dims(extinction, axis=0)
        spectra_container_out.rss_intensity = (spectra_container_out.rss_intensity
                                            * extinction)
        spectra_container_out.rss_variance = (spectra_container_out.rss_variance
                                           * extinction**2)
        self.record_correction(spectra_container_out, status='applied',
                            comment=comment)
        return spectra_container_out


# =============================================================================
# Differential Atmospheric Refraction
# =============================================================================
# TODO: refactor ADR by DAR. Create a class that performs this correction.
def get_adr(spectra_container : SpectraContainer, max_adr=0.5, pol_deg=2,
            plot=False):
    """Computes the ADR for a given DataContainer.
    
    This method computes the spatial shift as function of wavelength that a
    chromatic source light experiments due to the Atmospheric Differential
    Refraction (ADR).

    Parameters
    ----------
    spectra_container: :class:`SpectraContainer`
        Target SpectraContainer.
    max_adr : float, optional, default=0.5
        Maxium ADR correction expressed in arcseconds to prevent unreliable
        results when analyzing low-SNR data.
    pol_deg : int, optional, default=2
        Polynomial order to model the dependance of the spatial shift as function
        of wavelength.

    Returns
    -------
    - ra_polfit: np.poly1d
        Callable that returns the RA offset in arcseconds as function of wavelength.
    - dec_polfit: np.poly1d
        Callable that returns the DEC offset in arcseconds as function of wavelength.
    """
    # Centre of mass using multiple power of the intensity
    max_adr = check_unit(max_adr, u.arcsec)
    com = []
    for i in range(1, 5):
        com.append(spectra_container.get_centre_of_mass(power=i))
    com = np.array(com) << com[0][0].unit
    com -= np.nanmedian(com, axis=2)[:, :, np.newaxis]
    median_com = np.nanmedian(
        com, axis=0) - np.nanmedian(com, axis=(0, 2))[:, np.newaxis]
    median_com[np.abs(median_com) > max_adr] = np.nan

    finite_mask = np.isfinite(median_com[0])

    if finite_mask.any():
        p_x = np.polyfit(
            spectra_container.wavelength[finite_mask].to_value("angstrom"),
            median_com[0][finite_mask].to_value("arcsec"), deg=pol_deg)
        polfit_x = np.poly1d(p_x)(
            spectra_container.wavelength.to_value("angstrom")) << u.arcsec
    else:
        vprint("[ADR] ERROR: Could not compute ADR-x, all NaN")
        polfit_x = np.zeros(spectra_container.wavelength.size) << u.arcsec

    finite_mask = np.isfinite(median_com[1])
    if finite_mask.any():
        p_y = np.polyfit(
            spectra_container.wavelength[finite_mask].to_value("angstrom"),
            median_com[1][finite_mask].to_value("arcsec"), deg=pol_deg)
        polfit_y = np.poly1d(p_y)(
            spectra_container.wavelength.to_value("angstrom")) << u.arcsec
    else:
        vprint("[ADR] ERROR: Could not compute ADR-y, all NaN")
        polfit_y = np.zeros_like(spectra_container.wavelength)
    

    if plot:
        fig = plt.figure(figsize=(10, 5))
        ax = fig.add_subplot(121)
        ax.plot(spectra_container.wavelength, com[0, 0].to("arcsec"),
                label='COM', lw=0.7)
        ax.plot(spectra_container.wavelength, com[1, 0], label='COM2', lw=0.7)
        ax.plot(spectra_container.wavelength, com[2, 0], label='COM3', lw=0.7)
        ax.plot(spectra_container.wavelength, com[3, 0], label='COM4', lw=0.7)
        ax.plot(spectra_container.wavelength, median_com[0], c='k',
                label='Median', lw=0.7)
        ax.plot(spectra_container.wavelength, polfit_x, c='fuchsia',
                label=f'pol. fit (deg={pol_deg})', lw=0.7)
        ax.set_ylim(-max_adr, max_adr)
        ax.legend(ncol=2, bbox_to_anchor=(0.5, 1), loc='lower center')
        ax.set_ylabel(r'$\Delta RA$ (arcsec)')
        ax.set_xlabel(r'$\lambda$')

        ax = fig.add_subplot(122)
        ax.plot(spectra_container.wavelength, com[0, 1].to("arcsec"),
                label='COM', lw=0.7)
        ax.plot(spectra_container.wavelength, com[1, 1], label='COM2', lw=0.7)
        ax.plot(spectra_container.wavelength, com[2, 1], label='COM3', lw=0.7)
        ax.plot(spectra_container.wavelength, com[3, 1], label='COM4', lw=0.7)
        ax.plot(spectra_container.wavelength, median_com[1], c='k',
                label='Median', lw=0.7)
        ax.plot(spectra_container.wavelength, polfit_y, c='fuchsia',
                label=f'pol. fit (deg={pol_deg})', lw=0.7)
        ax.set_ylim(-max_adr, max_adr)
        ax.legend(ncol=2, bbox_to_anchor=(0.5, 1), loc='lower center')
        ax.set_ylabel(r'$\Delta DEC$ (arcsec)')
        ax.set_xlabel(r'$\lambda$')

        fig.subplots_adjust(wspace=0.3)
        plt.close(fig)
        return polfit_x, polfit_y, fig
    return polfit_x, polfit_y



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import os
from pykoala.ancillary import remove_units_dec

@remove_units_dec
def quick_plot(x, y,  xmin=None,xmax=None,ymin=None,ymax=None,percentile_min=2, percentile_max=98, extra_y = 0.1,
              ptitle = None, xlabel = None, ylabel = None, label="",
              #ptitle="Pretty plot", xlabel="Wavelength [$\mathrm{\AA}$]", ylabel="", 
              fcal=None, 
              psym=None, color=None, alpha=None, linewidth=1,  linestyle=None, markersize = 10,
              vlines=[], cvlines=[], svlines = [], wvlines=[], avlines=[], 
              hlines=[], chlines=[], shlines = [], whlines=[], ahlines=[],
              axvspan=None,  caxvspan = None,  aaxvspan = None,
              frameon = False, loc = 0, ncol = 6,   text=[],
              title_fontsize=12, label_axes_fontsize=10, axes_fontsize=10, tick_size=[5,1,2,1], axes_thickness =0,
              save_file=None, path=None, fig_size=9, warnings = True, verbose=True,
              show=True, statistics="", plot=True, **kwargs):
    
    
    
   """
   Plot this plot! An easy way of plotting plots in Python.
   
   Parameters
    ----------
    x,y : floats (default = None)
        Positional arguments for plotting the data
    xmin, xmax, ymin, ymax : floats (default = None)
        Plotting limits
    percentile_min : integer (default = 2)
        Lower bound percentile for filtering outliers in data
    percentile_max : integer (default = 98)
        Higher bound percentile for filtering outliers in data
    extra_y : float between 0 and 1
        Extra space up and down in y-axis, percentage of the plotting interval (ymax-ymin)
    ptitle : string (default = "Pretty plot")
        Title of the plot
    xlabel : string (default = "Wavelength [$\mathrm{\AA}$]")
        Label for x axis of the plot
    ylabel : string (default = "Flux [counts]")
        Label for y axis of the plot
    fcal : boolean (default = None)
        If True that means that flux is calibrated to CGS units, changes the y axis label to match
    psym : string or list of strings (default = "")
        Symbol marker, is given. If "" then it is a line.
    color : string (default = "blue")
        Color pallete of the plot. Default order is "red","blue","green","k","orange", "purple", "cyan", "lime"
    alpha : float or list of floats (default = 1)
        Opacity of the graph, 1 is opaque, 0 is transparent
    linewidth: float or list of floats (default=1)
        Linewidth of each line or marker
    linestyle:  string or list of strings (default="-")
        Style of the line
    markersize: float or list of floats (default = 10)
        Size of the marker
    vlines : list of floats (default = [])
        Draws vertical lines at the specified x positions
    hlines : list of floats (default = [])
        Draws horizontal lines at the specified y positions
    chlines : list of strings (default = [])
        Color of the horizontal lines
    cvlines : list of strings (default = [])
        Color of the vertical lines
    axvspan : list of floats (default = [[0,0]])
        Shades the region between the x positions specified
    caxvspan : list of strings (default = [])
        Color of the shaded regions     
    hwidth: float (default =1)
        thickness of horizontal lines
    vwidth: float (default =1)
        thickness of vertical lines
    frameon : boolean (default = False)
        Display the frame of the legend section
    loc : string or pair of floats or integer (default = 0)
        Location of the legend in pixels. See matplotlib.pyplot.legend() documentation
    ncol : integer (default = 6)
        Number of columns of the legend
    label : string or list of strings (default = "")
        Specify labels for the graphs
    title_fontsize: float (default=12)
        Size of the font of the title
    label_axes_fontsize: float (default=10)
        Size of the font of the label of the axes (e.g. Wavelength or Flux)
    axes_fontsize: float (default=10)
        Size of the font of the axes (e.g. 5000, 6000, ....)
    tick_size: list of 4 floats (default=[5,1,2,1])
        [length_major_tick_axes, thickness_major_tick_axes, length_minor_tick_axes, thickness_minor_tick_axes]   
        For defining the length and the thickness of both the major and minor ticks
    axes_thickness: float (default=0)
        Thickness of the axes
    save_file : string (default = none)
        Specify path and filename to save the plot
    path: string  (default = "")
        path of the file to be saved
    fig_size : float (default = 12)
        Size of the figure
    warnings : boolean (default = True)
        Print the warnings in the console if something works incorrectly or might require attention 
    show : boolean (default = True)
        Show the plot
    statistics : boolean (default = False)
        Print statistics of the data in the console
       
   """
   
   if color is None: color = "blue"
   if linestyle is None: linestyle ="-"
      
   if fig_size == "big":
       fig_size=20
       label_axes_fontsize=20 
       axes_fontsize=15
       title_fontsize=22
       tick_size=[10,1,5,1]
       axes_thickness =3
       if wvlines is None and vlines is not None: wvlines = [2] * len(vlines)
       if whlines is None and hlines is not None: whlines = [2] * len(hlines)

   if fig_size in ["very_big", "verybig", "vbig"]:
       fig_size=35
       label_axes_fontsize=30 
       axes_fontsize=25
       title_fontsize=28 
       tick_size=[15,2,8,2]
       axes_thickness =3
       if wvlines is None and vlines is not None: wvlines = [4] * len(vlines)
       if whlines is None and hlines is not None: whlines = [4] * len(hlines)

   if fig_size not in ["C","c","continue","Continue"]:
       if fig_size != 0 : 
           if type(fig_size) == list:
               plt.figure(figsize=(fig_size[0], fig_size[1]))
           else:     
               plt.figure(figsize=(fig_size, fig_size/2.5))
   
   if np.isscalar(x[0]) :
       xx=[]
       for i in range(len(y)):
           xx.append(x)
   else:
       xx=x

   if xmin is None : xmin = np.nanmin(xx[0])
   if xmax is None : xmax = np.nanmax(xx[0])  
   
   alpha_=alpha
   label_=label
   linewidth_=linewidth
   markersize_=markersize
   linestyle_=linestyle

   n_plots=len(y)
       
   if np.isscalar(y[0]) ==  False:
               
       if alpha_ is None:
           alpha =[0.5]*n_plots
       elif np.isscalar(alpha):
           alpha=[alpha_]*n_plots
       if psym is None: psym=[None]*n_plots
       if np.isscalar(label): label=[label_]*n_plots
       if np.isscalar(linewidth): linewidth=[linewidth_]*n_plots
       if np.isscalar(markersize):markersize=[markersize_]*n_plots
       if np.isscalar(linestyle): linestyle=[linestyle_]*n_plots
       if color == "blue" : 
           color_list = ["red","blue","green","k","orange", "purple", "cyan", "lime", "plum","teal","lightcoral", "navy"]
           color = []
           j=0
           for i in range(n_plots):
               color.append(color_list[j])
               j = j+1
               if j == 12 : j = 0
       if ymax is None: 
           y_max_list = []
       if ymin is None: y_min_list = []
              
       if fcal is None:
           if np.nanmedian(np.abs(y[0])) < 1E-10:
               fcal = True
               if np.nanmedian(y[0]) < 1E-20 and np.var(y[0]) > 0 : fcal = False
       for i in range(len(y)):
           #print("color=",color,"\n alpha=", alpha,"\n label=", label, "linewidth=", linewidth, "linestyle=", linestyle)
           if psym[i] is None:
             plt.plot(xx[i],y[i], color=color[i], alpha=alpha[i], label=label[i], linewidth=linewidth[i], linestyle=linestyle[i])
           else: 
               plt.plot(xx[i],y[i], psym[i], color=color[i], alpha=alpha[i], label=label[i], mew=linewidth[i], markersize=markersize[i])

           if ymax is None:
                    y_max_ = []                
                    y_max_.extend((y[i][j]) for j in range(len(xx[i])) if (xx[i][j] > xmin and xx[i][j] < xmax) )  
                    y_max_list.append(np.nanpercentile(y_max_, percentile_max))
           if ymin is None:
                    y_min_ = []                
                    y_min_.extend((y[i][j]) for j in range(len(xx[i])) if (xx[i][j] > xmin and xx[i][j] < xmax) )  
                    y_min_list.append(np.nanpercentile(y_min_, percentile_min))
       if ymax is None :
           ymax = np.nanmax(y_max_list)
       if ymin is None:
           ymin = np.nanmin(y_min_list)
   else:
       if alpha is None: alpha=1
       if statistics == None : statistics = True
       if fcal is None:
           if np.nanmedian(np.abs(y)) < 1E-10 : 
               fcal= True 
               if np.nanmedian(np.abs(y)) < 1E-20 and np.nanvar(np.abs(y)) > 0 : fcal = False
       if psym is None:
             plt.plot(x,y, color=color, alpha=alpha,linewidth=linewidth,  linestyle=linestyle)
       else:
           plt.plot(x,y, psym, color=color, alpha=alpha, mew=linewidth, markersize=markersize)
       if ymin is None :
           y_min_ = []                
           y_min_.extend((y[j]) for j in range(len(x)) if (x[j] > xmin and x[j] < xmax) )  
           ymin = np.nanpercentile(y_min_, percentile_min)
       if ymax is None :
           y_max_ = []                
           y_max_.extend((y[j]) for j in range(len(x)) if (x[j] > xmin and x[j] < xmax) )  
           ymax = np.nanpercentile(y_max_, percentile_max)    
       

   # Provide extra range up and down
   if extra_y is not None:
       interval = (ymax - ymin)
       ymin = ymin - interval * extra_y
       ymax = ymax + interval * extra_y
   
   # Set limits
   plt.xlim(xmin,xmax)
   plt.ylim(ymin,ymax)
   
   # Set title
   if ptitle is None: ptitle="Pretty plot"
   try:   
       plt.title(ptitle, fontsize=title_fontsize)
   except Exception:
       if warnings : print("  WARNING: Something failed when including the title of the plot")
   
   plt.minorticks_on()
   if xlabel is None: xlabel="Wavelength [$\mathrm{\AA}$]"
   plt.xlabel(xlabel, fontsize=label_axes_fontsize)
   #plt.xticks(rotation=90)
   plt.tick_params('both', length=tick_size[0], width=tick_size[1], which='major')
   plt.tick_params('both', length=tick_size[2], width=tick_size[3], which='minor')
   plt.tick_params(labelsize=axes_fontsize)
   plt.axhline(y=ymin,linewidth=axes_thickness, color="k")     # These 4 are for making the axes thicker, it works but it is not ideal...
   plt.axvline(x=xmin,linewidth=axes_thickness, color="k")    
   plt.axhline(y=ymax,linewidth=axes_thickness, color="k")    
   plt.axvline(x=xmax,linewidth=axes_thickness, color="k")    
   
   if ylabel is None: 
       if fcal: 
           ylabel = "Flux [ erg cm$^{-2}$ s$^{-1}$ $\mathrm{\AA}^{-1}$ ]"
       else:         
           ylabel = "Flux [counts]"  
 
   plt.ylabel(ylabel, fontsize=label_axes_fontsize)
   if label_ != "" : 
       plt.legend(frameon=frameon, loc=loc, ncol=ncol)
   
   # Plot horizontal lines
   if len(shlines) != len(hlines):
       for i in range(len(hlines)-len(shlines)):
           shlines.append("--")  
   if len(chlines) != len(hlines):
       for i in range(len(hlines)-len(chlines)):
           chlines.append("k")  
   if len(whlines) != len(hlines):
       for i in range(len(hlines)-len(whlines)):
           whlines.append(1)
   if len(ahlines) != len(hlines):
       for i in range(len(hlines)-len(ahlines)):
           if chlines[i] != "k":
               ahlines.append(0.8)
           else:
               ahlines.append(0.3)
   for i in range(len(hlines)):        
       plt.axhline(y=hlines[i], color=chlines[i], linestyle=shlines[i], alpha=ahlines[i], linewidth=whlines[i])
    
   # Plot vertical lines 
   if len(svlines) != len(vlines):
       for i in range(len(vlines)-len(svlines)):
           svlines.append("--")  
   if len(cvlines) != len(vlines):
       for i in range(len(vlines)-len(cvlines)):
           cvlines.append("k") 
   if len(wvlines) != len(vlines):
       for i in range(len(vlines)-len(wvlines)):
           wvlines.append(1)
   if len(avlines) != len(vlines):
       for i in range(len(vlines)-len(avlines)):
           if cvlines[i] != "k":
               avlines.append(0.8)
           else:
               avlines.append(0.3)
   for i in range(len(vlines)): 
       plt.axvline(x=vlines[i], color=cvlines[i], linestyle=svlines[i], alpha=avlines[i], linewidth=wvlines[i])
    
   # Plot horizontal ranges
   if axvspan is not None:
       if caxvspan is None:  caxvspan = ["orange" for item in axvspan]
       if aaxvspan is None:  aaxvspan = [0.15 for item in axvspan]
       for i in range(len(axvspan)):
           plt.axvspan(axvspan[i][0], axvspan[i][1], facecolor=caxvspan[i], alpha=aaxvspan[i], zorder=3)   
           
   if len(text)  > 0:
        for i in range(len(text)):
            plt.text(text[i][0],text[i][1], text[i][2], size=axes_fontsize)
                    
   if save_file == None:
       if show: 
           plt.show()
           plt.close() 
   else:
       if path != None : save_file=os.path.join(path, save_file) 
       plt.savefig(save_file,  dpi=300, bbox_inches='tight')
       plt.close() 
       if verbose: print("  Figure saved in file",save_file)
   
   if statistics == None: statistics=False
   if statistics:
       if np.isscalar(y[0]) : 
           basic_statistics(y, x=x, xmin=xmin, xmax=xmax)
       else:    
           for i in range(len(y)):
               basic_statistics(y[i], x=xx[i], xmin=xmin, xmax=xmax)



def basic_statistics(y, x=None, xmin=None, xmax=None, return_data=False, verbose = True):
    """
    Provides basic statistics: min, median, max, std, rms, and snr"
    """    
    if x is None:
        y_ = y
    else:          
        y_ = []      
        if xmin is None : xmin = x[0]
        if xmax is None : xmax = x[-1]          
        y_.extend((y[j]) for j in range(len(x)) if (x[j] > xmin and x[j] < xmax) )  
     
    median_value = np.nanmedian(y_)
    min_value = np.nanmin(y_)
    max_value = np.nanmax(y_)
    
    n_ = len(y_)
    mean_ = np.nanmean(y_)
    var_ = np.nanvar(y_)

    std = np.sqrt(var_)
    ave_ = np.nanmean(y_)
    disp_ =  max_value - min_value
    
    rms_v = ((y_ - mean_) / disp_ ) **2
    rms = disp_ * np.sqrt(np.nansum(rms_v)/ (n_-1))
    snr = ave_ / rms
    
    if verbose:
        print("  min_value  = {}, median value = {}, max_value = {}".format(min_value,median_value,max_value))
        print("  standard deviation = {}, rms = {}, snr = {}".format(std, rms, snr))
    
    if return_data : return min_value,median_value,max_value,std, rms, snr


def minimize_error(error_list, max_std_to_median_value = 0.25, verbose = False):
    """
    Providing a list of errors, check if they are similar within an intervale and if they are minimize error

    Parameters
    ----------
    error_list : list of floats
        list with errors.
    max_std_to_median_value : float, optional
        Max value of the STD/MEDIAN value to apply minimization of errors. The default is 0.25.
    verbose : Boolean, optional
        Print the errors in screen. The default is False.

    Returns
    -------
    error : float
        median or minimized error.

    """
    stat_error_list=basic_statistics(error_list, return_data=True, verbose=False)
    optimized = False
    if stat_error_list[3]/stat_error_list[1] < max_std_to_median_value:
        # Only 2 values of errors are combined: the MEDIAN value and the MAX value
        error = 1/stat_error_list[1]**2    +   1/np.nanmax(error_list)**2  
        # error=0                            # This makes the error too low!!!
        # for i in range(len(error_list)):  
        #     #eo2=(1/ea1**2 + 1/ea2**2 + 1/ea3**2 + 1/ea4**2)**(-0.5)
        #     error = error + 1 / error_list[i]**2 
        #     #print (1/np.sqrt(error))
        error = 1/np.sqrt(error)
        optimized = True
    else:
        error = stat_error_list[1]
    if verbose: print("Error list provided:",error_list,"\nEstimated error =",error, ", Optimized = ",optimized)
    return error


"""
Main exceptions used to trace errors within pykoala.
"""


class NoneAttrError(Exception):
    """
    Exception raised when trying to access to an attributed not provided.
    """
    def __init__(self, attr_name):
        self.message = ("Attribute {} not provided (None)".format(attr_name))
        super().__init__(self.message)


class ClassError(Exception):
    """
        Exception raised when providing wrong data
    """

    def __init__(self, expected_classes, input_value):
        expected_classes_string = ' {}\n' * len(expected_classes)
        self.message = ("Input object class {} ".format(input_value) +
                        "does not match expected classes:\n" + expected_classes_string.format(
                        *expected_classes))
        super().__init__(self.message)


class MaskError(Exception):
    """
    Exceptions raised for errors during data masking.
    """


class MaskBitError(MaskError):
    """
    Exception raised when providing wrong bit mask value
    """
    def __init__(self, accepted_values, input_value):
        self.message = "Mask value {} does not correspond default: {}"\
            .format(accepted_values, ', '.join(str(i) for i in input_value))
        super().__init__(self.message)


class CorrectionClassError(Exception):
    """
    Exception class raised for CorrectionBase classes
    """
    def __init__(self, target_class, wrong_target):
        self.message = "ERROR: Target class {} does not match input {}".format(target_class, wrong_target)
        super().__init__(self.message)


class TelluricError(Exception):
    """
    Parent exception class raised during telluric correction
    """


class TelluricNoFileError(TelluricError):
    """
    Parent exception class raised during telluric correction
    """
    def __init__(self):
        self.message = "ERROR: Not RSS nor Telluric file provided"
        super().__init__(self.message)


class FitError(Exception):
    """
    Parent exception class raised during telluric correction
    """
    def __init__(self):
        self.message = "Unsuccessful fit."
        super().__init__(self.message)


class CalibrationError(Exception):
    """
        Exception class raised during flux calibration
        """

    def __init__(self):
        self.message = "Data Container already flux calibrated."
        super().__init__(self.message)



"""
Astrometry-related corrections.
"""

# =============================================================================
# Basics packages
# =============================================================================
import numpy as np
# =============================================================================
# Astropy and associated packages
# =============================================================================
from astropy import units as u
from astropy import constants
from astropy.coordinates import SkyCoord
from photutils.centroids import centroid_2dg, centroid_com
from astropy import units as u

# =============================================================================
# KOALA packages
# =============================================================================
from pykoala import vprint
from pykoala.corrections.correction import CorrectionBase, CorrectionOffset
from pykoala.data_container import RSS, Cube
from pykoala.cubing import make_dummy_cube_from_rss
from pykoala.ancillary import interpolate_image_nonfinite
from pykoala.plotting.utils import qc_registration_centroids
from pykoala import photometry


class AstrometryOffsetCorrection(CorrectionBase):
    """Astrometry Offset correction class.

    This class accounts for relative astrometric offsets.

    Attributes
    ----------
    offset : AstrometryOffset
        Fibre astrometric offset.
    verbose: bool
        False by default.
    
    See also
    --------
    :class:`CorrectionBase`

    """
    name = "AstrometryOffsetCorrection"
    offset = None
    verbose = False

    def __init__(self, offset_path=None, offset=None, **correction_kwargs):
        super().__init__(**correction_kwargs)

        self.path = offset_path
        self.offset = offset

    @classmethod
    def from_fits(cls, path):
        """Initialise a AstrometryOffsetCorrection from an input FITS file.
        
        Parameters
        ----------
        path : str
            Path to the FITS file containing the offset data.

        Returns
        -------
        offset_correction : :class:`WavelengthCorrection`
            A :class:`WavelengthCorrection` initialised with the input data.
        """
        return cls(offset=CorrectionOffset.from_fits(path=path),
                   offset_path=path)

    @classmethod
    def from_external_image(cls, data_container, external_image, filter_name,
                            **crosscorr_kwargs):
        """Use a reference external image to compute the offset.
        
        Description
        -----------
        Estimate an astrometric offsets by cross-correlating a reference image with
        the DataContainer using mulitple aperture photometry.

        Parameters
        ----------
        data_container : :class:`pykoala.data_container.DataContainer`
            Target DataContainer.
        external_image : dict
            A dictionary containing the ``ccd`` of a reference image,
            corresponding to an instance of astropy.nddata.CCDData, and the
            pixels size ("pix_size") expressed in arcseconds.
        filter_name : str
            Name of the filter passband associated to the external image.
        crosscorr_kwargs : 
            Additional arguments to be passed to :func:`crosscorrelate_im_apertures`.

        Returns
        -------
        astrometry_offset_correction : class:`AstrometryOffsetCorrection`
            The resulting correction.
        results : dict
            Dictionary containing the results of the cross-correlation.
        """
        # Compute the synthetic photometry associated to the DataContainer
        dc_photometry = photometry.get_dc_aperture_flux(
            data_container, filter_name)
        # Only include valid (finite-valued) apertures
        vprint("Computing astrometric offsets")
        results = photometry.crosscorrelate_im_apertures(
            dc_photometry['aperture_flux'][dc_photometry['aperture_mask']],
            dc_photometry['coordinates'][dc_photometry['aperture_mask']],
            external_image, **crosscorr_kwargs)
        # Make a QC plot with the resulting solution
        fig = photometry.make_plot_astrometry_offset(
            data_container, dc_photometry['synth_photo'],
            external_image, results)
        results['offset_fig'] = fig

        return cls(offset=CorrectionOffset(offset_data=results['offset_min'],
                   offset_error=np.full_like(results['offset_min'],
                                             fill_value=np.nan))), results

    def apply(self, data_container):
        """Apply an astrometric offset correction to a DataContainer.

        Parameters
        ----------
        data_container : :class:`pykoala.data_container.DataContainer`
            DataContainer to be corrected.

        Returns
        -------
        dc_corrected : :class:`pykoala.data_container.DataContainer`
            Corrected copy of the input DC.
        """
        self.vprint(
            f"Applying astrometry offset correction to DC (RA, DEC): {self.offset.offset_data}")
        dc_out = data_container.copy()
        dc_out.update_coordinates(offset=self.offset.offset_data)
        self.record_correction(dc_out, status='applied')
        return dc_out


# TODO: Turn into a child of AstrometryOffsetCorrection
class AstrometryCorrection(CorrectionBase):
    """Perform astrometry-related corrections on DataContainers
    
    This class applies astrometry corrections to input DataContainers based
    on different kind of quantities.

    -   RSS offset correction
        Corrects the positions of all fibres using an offset (fib_ra = fib_ra + offset)
    -   Cube offset correction:
        Update the WCS central value according to a new pixel mapping.
    """
    name = 'Astrometry'
    verbose = True

    def __init__(self, **correction_args) -> None:
        super().__init__(**correction_args)

    def register_centroids(self, data_set, object_name=None, qc_plot=False, **centroid_args):
        """Register a collection of DataContainers.

        Description
        -----------
        The registration is performed by means of the function
        ``pykoala.corrections.astrometry.find_centroid_in_dc``. The first DataContainer
        will be used as reference, and the offset position of the star will be computed
        in the rest of DataContainers.

        Parameters
        ----------
        data_set: list
            Collection of DataContainers.
        object_name: str, default=None
            Database querable name for absolute astrometry calibration.
        qc_plot: bool, default=False
            If true, a QC plot will be generated by calling
            ``pykoala.plotting.utils.qc_registration_centroids``.
        **centroid_args:
            Extra arguments passed to the `pykoala.corrections.astrometry.find_centroid_in_dc` function.
        
        Returns
        -------
        - offsets: list
            A list containing the offsets for each input DataContainer.
        - fig: matplotlib.pyplot.Figure
            If `qc_plot=True`, it returns a quality control plot. `None` otherwise.
        """
        if len(data_set) <= 1:
            raise ArithmeticError("Input number of sources must be larger than 1")

        if object_name is not None:
            try:
                reference_pos = SkyCoord.from_name(object_name)
            except Exception as e:
                self.vprint(e)
                reference_pos = find_centroid_in_dc(data_set[0], **centroid_args)
        else:
            centroid_args['full_output'] = False
            reference_pos = find_centroid_in_dc(data_set[0], **centroid_args)
        if qc_plot:
            centroid_args['full_output'] = True
            images_list = []
            wcs_list = []
            centroid_list = []

        self.vprint(f"Reference position: RA={reference_pos.ra}, "
                    + f"DEC={reference_pos.dec})")
        offsets = []
        for data in data_set:
            if qc_plot:
                cube, image, _, centroid_world = find_centroid_in_dc(
                    data, **centroid_args)
                images_list.append(image)
                wcs_list.append(cube.wcs.celestial)
                centroid_list.append(centroid_world)                
            else:
                centroid_world = find_centroid_in_dc(data, **centroid_args)

            offset = [reference_pos.ra - centroid_world.ra, reference_pos.dec - centroid_world.dec]
            offsets.append(offset)
            self.vprint("Offset found (ra, dec): {}{} (arcsec)"
                        .format(offset[0].to_value('arcsec'),
                                offset[1].to_value('arcsec'))
                        )
        if qc_plot:
            fig = qc_registration_centroids(images_list, wcs_list,
                                            offsets, reference_pos)
            return offsets, fig
        else:
            return offsets, None

    def register_crosscorr(self, data_set, wave_range=None,
                           oversample=100, quick_cube_pix_size=0.3,
                           qc_plot=False):
        """Register a set of DataContainers using image cross-correlation.

        Parameters
        ----------
        data_set : :class:`pykoala.data_container.SpectraContainer`
            Set of SpectraContainers to cross-correlate.
        wave_range : list or tupla, default=None
            If provided, use the input wavelength range to create a white image.
        oversample: int, default=100
            Oversampling factor used during cross-correlation.
            See :func:`cross_correlate_images`.
        quick_cube_pix_size: float, default=0.3
            Pixels size of the data cube built to create the white image. This is
            only relevant when the SpectraContainer corresponds to a :class:`pykoala.rss.RSS`.
        qc_plot : bool, default=False
            If true, creates a quality control plot with the results of the
            cross-correlation match. Otherwise ``fig`` will be ``None``.

        Returns
        -------
        offsets: list of tuplas
            List of offsets in RA/DEC for every input SpectraContainer with respect
            to the first element of ``data_set``.
        fig: matplotlib.pyplot.Figure or None
            If ``qc_plot=True``, corresponds to a quality control figure containing
            the results of the cross-correlation. Otherwise is ``None``.
        """
        self.vprint("Performing image cross-correlation")
        images = []
        wcs = []
        offsets = [[0 * u.deg, 0 * u.deg]]

        for data_container in data_set:
            if data_container.__class__ is RSS:
                cube = make_dummy_cube_from_rss(data_container, quick_cube_pix_size)
                image = cube.get_white_image(wave_range=wave_range, s_clip=3.0)
            elif data_container.__class__ is Cube:
                cube = data_container
                image = cube.get_white_image(wave_range=wave_range, s_clip=3.0)
                image /= np.nansum(image)
            # Remove NaN values using N-Neighbours interpolation
            image = interpolate_image_nonfinite(image)
            # Convert the quantity into an array
            images.append(image)
            wcs.append(cube.wcs.celestial)
        # Perform the cross-correlation
        results = cross_correlate_images(images, oversample=oversample)
        for i in range(len(results)):
            pixels_shift = results[i][0]
            moving_origin = wcs[i + 1].pixel_to_world(0, 0)
            # TODO: check axis consistency
            reference_origin = wcs[0].pixel_to_world(0 - pixels_shift[1], 0 - pixels_shift[0])
            offsets.append([moving_origin.ra - reference_origin.ra,
                            moving_origin.dec - reference_origin.dec])
        if qc_plot:
            fig = qc_registration_centroids(images, wcs, offsets,
                                            ref_pos=wcs[i + 1].pixel_to_world(
                                                images[0].shape[1] / 2, images[0].shape[0] / 2))
            return offsets, fig
        else:
            return offsets, None
    
    def apply(self, data_container, offset):
        """Apply an astrometrical offset to a data container.
        
        Parameters
        ----------
        data_container : :class:`DataContainer`
            DataContainer to be corrected.
        offset : tuple

        Returns
        -------
        corrected_data_container : :class:`DataContainer`
            Copy of the input DataContainer including the astrometric correction.
        """
        if data_container.__class__ is RSS:
            self.vprint("Applying correction to RSS")

            data_container.info['fib_ra'] += offset[0]
            data_container.info['fib_dec'] += offset[1]

            self.record_correction(
                data_container, status='applied',
                offset=f"{offset[0].to('arcsec')}, {offset[1].to('arcsec')} arcsec")
        elif data_container.__class__ is Cube:
            # TODO: modify the WCS
            pass


def find_centroid_in_dc(data_container, wave_range=None,
                        centroider='com', com_power=1.0,
                        quick_cube_pix_size=0.5, subbox=None,
                        full_output=False):
    """Find the the centre of light in a DataContainer.

    Description
    -----------

    This method finds the centre of light on an image created from a
    DataContainer (RSS or Cube). The position is computed either by using
    the first moment or a gaussian kernel of some power of the light distribution.
    For RSS data, a previous step consists of building a new Cube.

    Parameters
    ----------
    data_container: DataContainer
        Data to extract the centroid.
    wave_range: list, default=None
        Wavelength range to be used to make the white image defined by the
        initial and final wavelengths.
    centroider: string, default='com'
        Keyword describing the centroider method to use. 'com' corresponds to
        the first moment of the light distribution, while 'gauss' uses a 2D gaussian
        kernel.
    com_power: float, default=1.0
        Power of the light distribution to weight the centroid position. The centroider
        function will compute the position of ``intensity**com_power``.
    quick_cube_pix_size: float, default=0.3 arcsec
        Only used when the DataContainer is an RSS. Pixel size in arcsec used to interpolate
        the RSS into a Cube.
    subbox: list, default=None
        Subbox containing the array indices of the image to select in the form:
        [[row_in, row_end], [col_in, col_end]].

    Returns
    -------
    centroid_world: astropy.coordinates.SkyCoord)
        Sky position of the centroid.
    """
    if centroider == 'com':
        centroider = centroid_com
    elif centroider == 'gauss':
        centroider = centroid_2dg

    if subbox is None:
        subbox = [[None, None], [None, None]]
    if isinstance(data_container, RSS):
        vprint(
            "[Registration]  Data provided in RSS format --> creating a dummy datacube")
        cube = make_dummy_cube_from_rss(data_container, quick_cube_pix_size)
        image = cube.get_white_image(wave_range=wave_range, s_clip=3.0)
    elif isinstance(data_container, Cube):
        cube = data_container
        image = cube.get_white_image(wave_range=wave_range, s_clip=3.0)
        image /= np.nansum(image)
    else:
        raise TypeError("Input DC must be an instance of Cube or RSS")
    # Select a subbox
    image = image[subbox[0][0]:subbox[0][1],
                  subbox[1][0]: subbox[1][1]]
    # Mask bad values
    # image = interpolate_image_nonfinite(image)
    centroider_mask = ~ np.isfinite(image)
    # Find centroid
    centroid_pixel  = np.array(centroider(image.value**com_power, centroider_mask))
    centroid_world = cube.wcs.celestial.pixel_to_world(*centroid_pixel)
    if not full_output:
        return centroid_world
    else:
        return cube, image, centroid_pixel, centroid_world 


def cross_correlate_images(list_of_images, oversample=100):
    """Compute a shift between several images using cross-correlation.
    
    Description
    -----------

    Apply the :func:`skimage.registration.phase_cross_correlation` method to find 
    the shift of a list of images with respect to the first one.

    Parameters
    ----------
    list_of_images : iterable
        Collection of images with same dimensions.
    oversample: int, default=100
        Oversampling factor used for the cross-correlation.
        See :func:`skimage.registration.phase_cross_correlation`.

    Returns
    -------
    results : list
        List containing the results of the cross correlation (shift, error, diffphase).
    """
    try:
        from skimage.registration import phase_cross_correlation
    except Exception:
        raise ImportError("For using the image crosscorrelation function you have to install scikit-image library")

    results = []
    for i in range(len(list_of_images) - 1):
        # The shift ordering is consistent with the input image shape
        shift, error, diffphase = phase_cross_correlation(
            list_of_images[0], list_of_images[i+1],
            upsample_factor=oversample)
        results.append([shift, error, diffphase])
    return results

# TODO: implement this method in the correction class
# def register_interactive(data_set, quick_cube_pix_size=0.2, wave_range=None):
#     """
#     Fully manual registration via interactive plot
#     """
#     import matplotlib
#     from matplotlib import pyplot as plt

#     images = []

#     for data in data_set:
#         if type(data) is RSS:
#             vprint(
#                 "[Registration]  Data provided in RSS format --> creating a dummy datacube")
#             cube_size_arcsec = (
#                 data.info['fib_ra'].max(
#                 ) - data.info['fib_ra'].min(),
#                 data.info['fib_dec'].max()
#                 - data.info['fib_dec'].min())
#             x, y = np.meshgrid(
#                 np.arange(- cube_size_arcsec[1] / 2 + quick_cube_pix_size / 2,
#                           cube_size_arcsec[1] / 2 + quick_cube_pix_size / 2,
#                           quick_cube_pix_size),
#                 np.arange(- cube_size_arcsec[0] / 2 + quick_cube_pix_size / 2,
#                           cube_size_arcsec[0] / 2 + quick_cube_pix_size / 2,
#                           quick_cube_pix_size))
#             datacube = np.zeros((data.wavelength.size, *x.shape))
    
#             # Interpolate RSS to a datacube
#             datacube, _, _ = interpolate_rss(
#                 data, pixel_size_arcsec=quick_cube_pix_size,
#                 kernel_scale=1.,
#                 datacube=datacube)
#             image = make_white_image_from_array(datacube, data.wavelength,
#                                                 wave_range=wave_range, s_clip=3.0)
#             image_pix_size = quick_cube_pix_size

#         elif type(data) is Cube:
#             image = data.get_white_image(wave_range=wave_range, s_clip=3.0)
#             image_pix_size = data.info['pixel_size_arcsec']

#         # Mask NaN values
#         image = interpolate_image_nonfinite(image)
#         images.append(image)

#     ### Start the interactive GUI ###
#     # Close previus figures
#     plt.close()
#     plt.ion()
#     centres = []

#     def mouse_event(event):
#         """..."""
#         centres.append([event.xdata / 3600, event.ydata / 3600])
#         vprint('x: {} and y: {}'.format(event.xdata, event.ydata))

#     n_rss = len(data_set)

#     fig = plt.figure()
#     cid = fig.canvas.mpl_connect('button_press_event', mouse_event)
#     for i in range(n_rss):
#         ax = fig.add_subplot(1, n_rss, i + 1)
#         ax.scatter(data_set[0].info["fib_ra"], data_set[0].info["fib_dec"],
#                    c=np.nanmedian(data_set[0].intensity_corrected, axis=1))
#     plt.show()
#     plt.ioff()
#     return centres

# Mr Krtxo \(ﾟ▽ﾟ)/



"""
Library of astronomical corrections applied to DataContainers.
"""
from .correction import CorrectionBase
from .astrometry import AstrometryCorrection
from .atmospheric_corrections import AtmosphericExtCorrection
from .flux_calibration import FluxCalibration

__all__ = ["CorrectionBase", "AstrometryCorrection", "AtmosphericExtCorrection",
           "FluxCalibration"]